'''
These are the things that are used when you `cytoplasm serve`.
'''

import os, cytoplasm, sys

# make this work in either Python 2.x or 3.x
if sys.version_info.major >= 3:
    from http.server import SimpleHTTPRequestHandler, HTTPServer
else:
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from BaseHTTPServer import HTTPServer

# keep track of when things were last built in this global variable
most_recent_time = 0

def serve(port, rebuild, event=None):
    "Serve the Cytoplasm site."
    # keep track of the most recently modified time in global variable most_recent_time
    global most_recent_time
    most_recent_time = most_recent(".")
    # rebuild the site first.
    cytoplasm.build()
    # change to the build directory, where things are to be served from.
    os.chdir(cytoplasm.configuration.get_config().build_dir)
    # use either SimpleHTTPRequestHandler or RebuildHandler, depending on whether rebuild is True.
    if rebuild: handler = RebuildHandler
    else: handler = SimpleHTTPRequestHandler
    # make a server with the handler and the port
    httpd = HTTPServer(('', port), handler)
    # serve!
    httpd.serve_forever()

def most_recent(directory):
    "Determine the most recent modified time in the source directory, ignoring dotfiles and _build."
    build_dir = cytoplasm.configuration.get_config().build_dir
    # get the candidate files:
    files = [f for f in os.listdir(directory) if f != build_dir and not f.startswith(".")]
    # get each of their times
    times = [os.stat(os.path.join(directory, f)).st_mtime for f in files]
    # the highest time here is the most recent; return that.
    return max(times)


class RebuildHandler(SimpleHTTPRequestHandler):
    def handle(self):
        "Handle a request and, if anything has changed, rebuild the site."
        # overwrite the handle method in SimpleHTTPRequestHandler with this.
        # declare most_recent_time global; we'll be changing it later.
        global most_recent_time
        # figure out what the most recent time edited is in the source directory
        new_recent = most_recent("..")
        # only build the site if the new most recent is more recent than the old one,
        # i.e. if one or more of the files has been edited.
        if new_recent > most_recent_time:
            # update most_recent_time
            most_recent_time = new_recent
            # Build the site from the source directory
            cytoplasm.build("..")
        # Call SimpleHTTPRequestHandler.handle(), so it can do stuff there too.
        SimpleHTTPRequestHandler.handle(self)




'''
These are some utilites used when writing and handling interpreters.
'''

import shutil
from cytoplasm import configuration
from cytoplasm.errors import InterpreterError

def interpreted_filename(file):
    "Intelligently return the interpreted filename of a file."
    # get the last suffix:
    suffix = file.split(".")[-1]
    # if the suffix is in interpreters.keys() the destination is everything but that last suffix
    if suffix in configuration.get_config().interpreters.keys():
        return ".".join(file.split(".")[:-1])
    # otherwise, it's the whole thing.
    else:
        return file

def SaveReturned(fn):
    '''Some potential interpreters, like Mako, don't give you an easy way to save to a destination.
    In these cases, simply use this function as a decorater.'''
    def InterpreterWithSave(source, destination, **kwargs):
        # under some circumstances, this should be able to write to file-like objects;
        # so if destination is a string, assume it's a path; otherwise, it's a file-like object
        if isinstance(destination, str):
            f = open(destination, "w")
        else:
            f = destination
        # pass **kwargs to this function.
        f.write(fn(source, **kwargs))
        f.close()
    return InterpreterWithSave

@SaveReturned
def default_interpreter(source, **kwargs):
    f = open(source)
    source_string = f.read()
    f.close()
    return source_string

def interpret(file, destination, **kwargs):
    "Interpret a file with an interpreter according to its suffix."
    # get the list of interpreters from the configuration
    interpreters = configuration.get_config().interpreters
    # figure out the suffix of the file, to use to determine which interpreter to use
    ending = file.split(".")[-1]
    interpreters.get(ending, default_interpreter)(file, destination, **kwargs)




'''
A nose tests suite for Cytoplasm. 
Run these by going into the cytoplasm package directory and running `nosetests`.
These work on the example sites in `cytoplasm/tests`.
'''

import os, imp, shutil
import nose, mako
import cytoplasm
from cytoplasm import server, configuration

# Figure out where the examples are, in the cytoplasm package.
_, cytoplasm_directory, _ = imp.find_module("cytoplasm")
# this is the directory
examples_directory = os.path.join(cytoplasm_directory, "tests")

# Infrastructure to use on the generalized tests.
class Base(object):
    "This is the base that all the test classes will inherit from."
    def __init__(self, directory):
        self.directory = directory
        self.build_dir = os.path.join(directory, "_build")
        self.configuration = configuration.get_config(directory)
    
    def delete_build_dir(self):
        "Delete the build directory."
        if os.path.exists(self.build_dir): shutil.rmtree(self.build_dir)

    def setup(self):
        "Delete all the files in the build directory already and build the site."
        self.delete_build_dir()
        cytoplasm.build(self.directory)

    def teardown(self):
        "Delete the build directory after all the tests are done."
        self.delete_build_dir()

    def basic_test(self):
        "Check for basic decency in a built cytoplasm site."
        # check that the build directory was created
        assert os.path.exists(self.build_dir)
        # check that the build directory is, in fact, a directory
        assert os.path.isdir(self.build_dir)

    def copy_html_test(self):
        "Test whether building a cytoplasm site correctly copies over the uninterpreted files."
        # a filter to tell whether files are html files and are not configuration files.
        filter = lambda x: x.endswith(".html") and not x.startswith("_")
        # the html files in the source directory:
        in_source_dir = [file for file in os.listdir(self.directory) if filter(file)]
        # the html files in the build dir:
        in_build_dir = [file for file in os.listdir(self.build_dir) if filter(file)]
        # If nothing bad has happenned, everything in in_source_dir should be in in_build_dir
        assert set(in_source_dir) <= set(in_build_dir)
        # Furthermore, the contents of each of these files should be the same.
        for source_path in in_source_dir:
            # open each of these files from their respective directories
            source_file = open(os.path.join(self.directory, source_path))
            built_file = open(os.path.join(self.build_dir, source_path))
            # assert that their contents are the same
            assert source_file.read() == built_file.read(), zip(in_source_dir, in_build_dir)
            # and then close each file
            source_file.close()
            built_file.close()

# The classes for different example sites:
# I'd love for this to be generator that acts as a class factory, but I can't even find if I can do
# it in nose. Sigh.

class TestEmpty(Base):
    "Test the empty site."
    def __init__(self):
        Base.__init__(self, os.path.join(examples_directory, "empty"))

class TestSomeHTML(Base):
    "Test the somehtml example site."
    def __init__(self):
        Base.__init__(self, os.path.join(examples_directory, "somehtml"))
    
    def interpreter_test(self):
        "Test that the interpreters basically work."
        # I'm not going to do any content testing, because that's mako's job.
        # Still, I'm going to test the files are copied over and interpreted.
        # for each .mako file in the source directory.
        for file in [file for file in os.listdir(self.directory) if file.endswith(".mako")]:
            # the corresponding file will be the same, but without the last dotted section.
            corresponding = ".".join(file.split(".")[:-1])
            # assert that there exists a corresponding file in the build directory:
            assert os.path.exists(os.path.join(self.build_dir, corresponding))

class TestControllers(Base):
    "Test the controllers example site."
    def __init__(self):
        Base.__init__(self, os.path.join(examples_directory, "controllers"))

    def basic_controller_test(self):
        # for each controller configured:
        for controller, [source_dir, build_dir] in self.configuration.controllers:
            # make sure the build directories were created.
            assert os.path.exists(os.path.join(self.directory, build_dir))

    def copier_controller_test(self):
        # get the controllers whose names are "copier"
        copiers = [c for c in self.configuration.controllers if c[0] == "copier"]
        for _, [source_dir, build_dir] in copiers:
            # for each of the files in the build directory...
            for file in os.listdir(os.path.join(self.directory, source_dir)):
                source_file = open(os.path.join(self.directory, source_dir, file))
                built_file = open(os.path.join(self.directory, source_dir, file))
                # check that they were copied correctly.
                assert source_file.read() == built_file.read()




'''
This module imports everything from the user's `_config.py` into its namespace.
So to get configuration of the current working directory, `import cytoplasm.configuration`.
'''

import os, imp
from cytoplasm.errors import CytoplasmError

# This a global variable to be shared and possibly changed across all of the cytoplasm modules.
source_dir = "."

def get_config(dir=None):
    # if there isn't a specified directory in the arguments, use the global source_dir.
    if dir == None: dir = source_dir
    # If the user has a file called _config.py, import that and return the module.
    # The user's _config.py should "from cytoplasm.defaults import *" if they want to use
    # some of the defaults.
    config_file = os.path.join(dir, "_config.py")
    if os.path.exists(config_file):
        imp.load_source("_config", config_file)
        import _config
        return _config
    else:
        raise CytoplasmError("You don't seem to have a configuration file at '%s'." %(source_dir))



'''
These are all the functions that are used when you `cytoplasm build`.
'''

import os, shutil
from cytoplasm import errors, interpreters, configuration
from cytoplasm.controllers import controllerclass

def copy_over(config):
    "Copy the files not beginning with '_' to the site directory"
    interpreter_dictionary = config.interpreters
    # All of the files and directories that don't begin with '_' or '.'
    to_copy = [file for file in os.listdir(".") if not file.startswith(("_", "."))]
    files = [file for file in to_copy if not os.path.isdir(file)]
    for file in files:
        destination = interpreters.interpreted_filename(file)
        # and pass the origin and destination to interpreters.interpret.
        interpreters.interpret(file, os.path.join(config.build_dir, destination))
    directories = [file for file in to_copy if os.path.isdir(file)]
    for dir in directories:
        if os.path.exists(os.path.join(config.build_dir, dir)):
            # if the directory exists, delete it.
            shutil.rmtree(os.path.join(config.build_dir, dir))
        # copy all the directories that don't start with "." or "_" completely.
        shutil.copytree(dir, os.path.join(config.build_dir, dir))

def build(dir="."):
    # Set the directory in cytoplasm.configuration, so everything gets the same configuration.
    configuration.source_dir = os.path.join(os.getcwd(), dir)
    # Get the configuration file...
    config = configuration.get_config()
    # save the current working directory, so we can return to it later.
    oldwd = os.getcwd()
    # and then change directories, to the source directory.
    os.chdir(configuration.source_dir)
    # Create the build directory, if it doesn't exist
    if not os.path.exists(config.build_dir):
        os.mkdir(config.build_dir)
    copy_over(config)
    for controller, arguments in config.controllers:
        # create a controller object of the class returned by controllerclass
        controller_object = controllerclass(controller)(*arguments)
        # call the controller object 
        controller_object()
    # change back to the old directory.
    os.chdir(oldwd)




'''
This is a default configuration file, that should be imported in the user's configuration
'''

import os, imp, sys
from cytoplasm import controllers

# Specify the build directory, where the built site will be copied to
build_dir = "_build"

from .interpreters import SaveReturned
# A dictionary where the key of a suffix is the function that should handle that file
# The function is given two arguments, the filename and where that file should be put 
# after it's finished
interpreters = {}

# This is a decorator that can be prepended to an interpreter function, so that you don't need
# to have the manually add each to the interpreters dictionary.
def Interpreter(*keys):
    # You can have as many suffixes as you want.
    def Add(fn):
        for suffix in keys:
            interpreters[suffix] = fn
    return Add

# An interpreter for mako, which will be prett essential both for "static" files and for controller
import mako.lookup, mako.template
# Mako should look for templates to include in the current directory.
# This should let you include things like "_templates/site.mako"
mako_lookup = mako.lookup.TemplateLookup(directories=['.'])
# Mako doesn't come with an easy built-in way to save the output to a certain file;
# this decorator does that.
@Interpreter("mako")
@SaveReturned
def mako_interpreter(file, **kwargs):
    # pass kwargs to the mako template
    page = mako.template.Template(filename=file, lookup=mako_lookup, input_encoding='utf-8')
    # this is dumb but it's the only way I can make it work.
    if sys.version_info.major == 2:
        # if it's python 2, use .encode('utf-8', 'replace')
        return page.render_unicode(**kwargs).encode('utf-8', 'replace')
    elif sys.version_info.major == 3:
        # otherwise, just render it...
        return page.render_unicode(**kwargs)

# This is a list of controllers to be used.
# Each item should be a 2-tuple containing:
# * The controller's module's name; i.e, the thing in _controllers
# * The list of arguments to use when instantiating the controller class; usually at least the 
#   controller's source directory and where files should be output to, in the build directory.
controllers = []




'''
These are the things used in making and using controllers.
'''

import imp, os

def controllerclass(name):
    "Given the name of a controller's module, return that controller's class."
    fp, pathname, description = imp.find_module(name, ["_controllers/"])
    try:
        controller = imp.load_module(name, fp, pathname, description)
        return controller.info["class"]
    finally:
        if fp != None: fp.close()

class Controller(object):
    "Controllers take data from some files, do stuff with it, and write it to the build directory."
    def __init__(self, data, destination, templates="_templates"):
        # take three arguments: the source directory, the destination directory, and, optionally,
        # a directory wherein the templates reside.
        self.data_directory = data
        self.destination_directory = destination
        self.templates_directory = templates
        # create the destination directory, if it doesn't exist
        if not os.path.exists(self.destination_directory): os.mkdir(self.destination_directory)
    
    def __call__(self):
        # do whatever needs to be done here...
        pass

    def template(self, name):
        "Given a name, return the name of the file in the templates directory that fits it."
        # list of files that fit:
        files = [f for f in os.listdir(self.templates_directory) if f.startswith(name + ".")]
        # return the first file that matches, with the directory prepended.
        return os.path.join(self.templates_directory, files[0])



from distutils.core import setup

setup(
    name = "Cytoplasm",
    version = "0.04.3",
    author = "startling",
    author_email = "tdixon51793@gmail.com",
    url = "http://cytoplasm.somethingsido.com",
    keywords = ["blogging", "site compiler", "blog compiler"],
    description = "A static, blog-aware website generator written in python.",
    packages = ['cytoplasm'],
    scripts = ['scripts/cytoplasm'],
    install_requires = ['Mako'],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary",
        ]
)
        
